using System;
using System.Drawing;
using System.Drawing.Imaging;
using System.Collections.Generic;
using System.Text;
using System.Runtime.InteropServices;
using System.IO;
using System.Diagnostics;

namespace csavi {
    class AVIManager {
        #region API錾
        static uint mmioFOURCC(char c0, char c1, char c2, char c3) {
            return (uint)((byte)c0 | ((byte)c1 << 8) | ((byte)c2 << 16) | ((byte)c3 << 24));
        }

        static readonly uint streamtypeVIDEO = mmioFOURCC('v', 'i', 'd', 's');
        static readonly uint streamtypeAUDIO = mmioFOURCC('a', 'u', 'd', 's');

        const int OF_READ = 0;
        const int OF_WRITE = 1;
        const int OF_CREATE = 4096;

        [StructLayout(LayoutKind.Sequential, Pack = 1)]
        struct AVISTREAMINFOW {
            public UInt32 fccType, fccHandler, dwFlags, dwCaps;
            public UInt16 wPriority, wLanguage;
            public UInt32 dwScale, dwRate,
                             dwStart, dwLength, dwInitialFrames, dwSuggestedBufferSize,
                             dwQuality, dwSampleSize, rect_left, rect_top,
                             rect_right, rect_bottom, dwEditCount, dwFormatChangeCount;
            public UInt16 szName0, szName1, szName2, szName3, szName4, szName5,
                             szName6, szName7, szName8, szName9, szName10, szName11,
                             szName12, szName13, szName14, szName15, szName16, szName17,
                             szName18, szName19, szName20, szName21, szName22, szName23,
                             szName24, szName25, szName26, szName27, szName28, szName29,
                             szName30, szName31, szName32, szName33, szName34, szName35,
                             szName36, szName37, szName38, szName39, szName40, szName41,
                             szName42, szName43, szName44, szName45, szName46, szName47,
                             szName48, szName49, szName50, szName51, szName52, szName53,
                             szName54, szName55, szName56, szName57, szName58, szName59,
                             szName60, szName61, szName62, szName63;
        }
        // vfw.h
        [StructLayout(LayoutKind.Sequential, Pack = 1)]
        struct AVICOMPRESSOPTIONS {
            public UInt32 fccType;
            public UInt32 fccHandler;
            public UInt32 dwKeyFrameEvery;  // only used with AVICOMRPESSF_KEYFRAMES
            public UInt32 dwQuality;
            public UInt32 dwBytesPerSecond; // only used with AVICOMPRESSF_DATARATE
            public UInt32 dwFlags;
            public IntPtr lpFormat;
            public UInt32 cbFormat;
            public IntPtr lpParms;
            public UInt32 cbParms;
            public UInt32 dwInterleaveEvery;
        }

        [StructLayout(LayoutKind.Sequential, Pack = 1)]
        struct BITMAPINFOHEADER {
            public UInt32 biSize;
            public Int32 biWidth;
            public Int32 biHeight;
            public Int16 biPlanes;
            public Int16 biBitCount;
            public UInt32 biCompression;
            public UInt32 biSizeImage;
            public Int32 biXPelsPerMeter;
            public Int32 biYPelsPerMeter;
            public UInt32 biClrUsed;
            public UInt32 biClrImportant;
        }

        [StructLayout(LayoutKind.Sequential, Pack = 1)]
        struct WAVEFORMATEX {
            public Int16 FormatTag;
            public Int16 Channels;
            public Int32 SamplesPerSec;
            public Int32 AvgBytesPerSec;
            public Int16 BlockAlign;
            public Int16 BitsPerSample;
            public Int16 Size;
        }


        [DllImport("avifil32.dll")]
        private static extern void AVIFileInit();

        [DllImport("avifil32.dll")]
        private static extern int AVIFileOpenW(ref int ptr_pfile, [MarshalAs(UnmanagedType.LPWStr)]string fileName, int flags, int dummy);

        [DllImport("avifil32.dll")]
        private static extern int AVIFileCreateStream(
          int ptr_pfile, out IntPtr ptr_ptr_avi, ref AVISTREAMINFOW ptr_streaminfo);

        [DllImport("avifil32.dll")]
        private static extern int AVIMakeCompressedStream(
          out IntPtr ppsCompressed, IntPtr aviStream, ref AVICOMPRESSOPTIONS ao, int dummy);

        [DllImport("avifil32.dll")]
        private static extern int AVIStreamSetFormat(
          IntPtr aviStream, Int32 lPos, ref BITMAPINFOHEADER lpFormat, Int32 cbFormat);

        [DllImport("avifil32.dll")]
        private static extern int AVIStreamSetFormat(
          IntPtr aviStream, Int32 lPos, ref WAVEFORMATEX lpFormat, Int32 cbFormat);

        [DllImport("avifil32.dll")]
        unsafe private static extern int AVISaveOptions(
          int hwnd, UInt32 flags, int nStreams, IntPtr* ptr_ptr_avi, AVICOMPRESSOPTIONS** ao);

        [DllImport("avifil32.dll")]
        private static extern int AVIStreamWrite(
          IntPtr aviStream, Int32 lStart, Int32 lSamples, IntPtr lpBuffer,
          Int32 cbBuffer, Int32 dwFlags, Int32 dummy1, Int32 dummy2);

        [DllImport("avifil32.dll")]
        private static extern int AVIStreamRelease(IntPtr aviStream);

        [DllImport("avifil32.dll")]
        private static extern int AVIFileRelease(int pfile);

        [DllImport("avifil32.dll")]
        private static extern void AVIFileExit();
        #endregion

        //pϐ
        int framerate_, width_, height_, stride_, pfile_, count_a_, count_v_;
        Bitmap bmp_;
        Graphics g_;
        Bitmap bb_bmp_;
        Graphics bb_g_;
        IntPtr pvs_, pas_;


        public Graphics Graphic { get { return g_; } }
        public Bitmap Bitmap { get { return bmp_; } }

        //t@CEt[[gETCYw肵ăt@CI[vA

        unsafe public Bitmap OpenFile(string filename, int framerate, int width, int height) {
            //ݒ̕ۑ
            framerate_ = framerate;
            width_ = width;
            height_ = height;
            count_a_ = count_v_ = 0;

            //pobt@Astride擾
            bmp_ = new Bitmap(width, height, PixelFormat.Format24bppRgb);

            bb_bmp_ = new Bitmap(width, height, PixelFormat.Format24bppRgb);
            BitmapData bmpDat = bb_bmp_.LockBits(new Rectangle(0, 0, width, height), ImageLockMode.ReadOnly, PixelFormat.Format24bppRgb);
            stride_ = bmpDat.Stride;
            bb_bmp_.UnlockBits(bmpDat);

            //APIĂяo
            AVIFileInit();

            int ret = AVIFileOpenW(ref pfile_, filename, 4097, 0);
            if (ret != 0)
                return null;

            #region 摜̃Xg[
            AVISTREAMINFOW asi_v = new AVISTREAMINFOW();
            asi_v.fccType = streamtypeVIDEO;//vids
            //asi_v.fccHandler = 808810089;
            asi_v.dwScale = 1;
            asi_v.dwRate = (uint)framerate;
            asi_v.dwSuggestedBufferSize = (uint)(height * stride_);
            asi_v.dwQuality = (uint)0xffffffff;
            asi_v.rect_bottom = (uint)height;
            asi_v.rect_right = (uint)width;
            asi_v.dwLength = (uint)(10 * framerate);

            ret = AVIFileCreateStream(pfile_, out pvs_, ref asi_v);
            if (ret != 0)
                return null;
            BITMAPINFOHEADER bi = new BITMAPINFOHEADER();
            bi.biSize = 40;
            bi.biWidth = width;
            bi.biHeight = height;
            bi.biPlanes = 1;
            bi.biBitCount = 24;
            bi.biSizeImage = (uint)(stride_ * height);
            ret = AVIStreamSetFormat(pvs_, 0, ref bi, 40);
            if (ret != 0)
                return null;
            g_ = Graphics.FromImage(bmp_);
            bb_g_ = Graphics.FromImage(bb_bmp_);

            #endregion

            #region ̃Xg[
            //44100HzA16bitAXeIŒ
            AVISTREAMINFOW asi_a = new AVISTREAMINFOW();
            asi_a.fccType = streamtypeAUDIO;//auds
            asi_a.dwScale = asi_a.dwSampleSize = 4;
            asi_a.dwRate = 44100 * 4;
            asi_a.dwQuality = (uint)0xffffffff;
            asi_a.dwLength = 5 * 44100;
            asi_a.dwInitialFrames = 44100;

            ret = AVIFileCreateStream(pfile_, out pas_, ref asi_a);
            if (ret != 0)
                return null;
            WAVEFORMATEX wf = new WAVEFORMATEX();
            wf.BlockAlign = 4;
            wf.AvgBytesPerSec = 44100 * 4;
            wf.Channels = 2;
            wf.SamplesPerSec = 44100;
            wf.FormatTag = 1;
            wf.BitsPerSample = 16;
            wf.Size = 0;
            ret = AVIStreamSetFormat(pas_, 0, ref wf, sizeof(WAVEFORMATEX));
            if (ret != 0)
                return null;
            #endregion

            //obt@pBitmapԂ
            return bmp_;
        }

        public void Close() {
            //摜щp̃Xg[ďI
            AVIStreamRelease(pvs_);
            AVIStreamRelease(pas_);
            pvs_ = pas_ = new IntPtr(0);
            if (pfile_ != 0) {
                AVIFileRelease(pfile_);
                pfile_ = 0;
            }

            AVIFileExit();
            g_.Dispose();
            bb_g_.Dispose();
            bb_bmp_.Dispose();
            bmp_.Dispose();

            bb_g_ = g_ = null;
            bb_bmp_ = bmp_ = null;
            
        }

        public void AddVideoFrame() {
            //݂̃rbg}bv̓et[ɒǉ
            AddVideoFrame(count_a_);
        }

        public void AddVideoFrame(int frame) {
            //t[ԍw肵Ēǉ
            count_a_ = frame;
            //㉺]
            bb_g_.DrawImage(bmp_, new Rectangle(0, height_-1, width_, -height_), 0, 0, width_, height_, GraphicsUnit.Pixel);

            BitmapData bd = bb_bmp_.LockBits(new Rectangle(0, 0, width_, height_), ImageLockMode.ReadOnly, PixelFormat.Format24bppRgb);
            AVIStreamWrite(pvs_, count_a_, 1, bd.Scan0, stride_ * height_, 16, 0, 0);
            bb_bmp_.UnlockBits(bd);
            count_a_++;
        }

        unsafe public void AddAudioFrame(Int16[] data) {
            //ǉA1bxǂ
            int frame = data.Length / 2;
            IntPtr raw = Marshal.AllocHGlobal(data.Length * 2);
            Marshal.Copy(data, 0, raw, data.Length);
            int ret = AVIStreamWrite(pas_, count_v_, frame, raw, 4 * frame, 0, 0, 0);
            Marshal.FreeHGlobal(raw);
            count_v_ += frame;
        }
    }
}
